package net.sergeych.mptools

import kotlin.random.Random

class SafeId {

    companion object {

        operator fun get(encoded: String): ByteArray = decode(encoded)
        operator fun get(source: ByteArray): String = encodeToGroups(source)

        val alphabet = "ABCDEFGHJKLMNPQRSTUWXYZ012345679"
//        val alphaValues = alphabet.mapIndexed { i, c -> c to i }.toMap()

        val codeTable = IntArray(128)

        init {
            if (alphabet.length != 32) throw Error("Alphabet should be 32 chars")
            codeTable.fill(-1)
            for ((i, x) in alphabet.withIndex()) {
                codeTable[x.code] = i
            }
            for ((i, x) in alphabet.lowercase().withIndex()) {
                codeTable[x.code] = i
            }
            codeTable['8'.code] = codeTable['B'.code]
            codeTable['I'.code] = codeTable['1'.code]
            codeTable['i'.code] = codeTable['1'.code]
            codeTable['O'.code] = codeTable['0'.code]
            codeTable['o'.code] = codeTable['0'.code]
            codeTable['V'.code] = codeTable['U'.code]
            codeTable['v'.code] = codeTable['U'.code]
        }

        /**
         * Create random string safeId that represents a byte array of specified length
         */
        fun random(bytesLength: Int): String {
            return SafeId[Random.nextBytes(bytesLength)]
        }

        fun encode(data: ByteArray): String {
            var buffer = 0
            var bits = 0

            val output = StringBuilder()

            fun emit(limit: Int = 5) {
                while (bits >= limit) {
                    output.append(alphabet[buffer and 0x1F])
                    bits -= 5
                    buffer = buffer shr 5
                }
            }

            for (sb in data) {
                emit()
                val i = sb.toUByte().toInt()
                buffer = (i shl bits) or buffer
                bits += 8
            }
            emit(0)
            return output.toString()
        }

        /**
         * Encode [data] to `SafeId` string of groups up to [n] characters separated by '-'
         */
        fun encodeToGroups(data: ByteArray, n: Int = 4): String {
            val x = encode(data)
            var i = 0;
            val result = StringBuilder()
            while (i < x.length) {
                if (i != 0) result.append('-')
                var max = i + n
                if (max > x.length) max = x.length
                result.append(x.substring(i until max))
                i += n
            }
            return result.toString()
        }

        /**
         * Decode `safeId` string, ignoring any group separation characters and spaces, and correcting possibly
         * mistypes of similar looking characters, like 'I' instead of '1'.
         * @throws IllegalArgumentException if source contains not correctable illegal characters
         */
        fun decode(source: String): ByteArray {
            val result = mutableListOf<Byte>()
            var buffer = 0
            var bits = 0
            for (ch in source) {
                val i = when (ch) {
                    ' ', '-', '\t' -> continue
                    else -> {
                        val code = ch.code
                        if (code < 0 || code >= codeTable.size)
                            throw IllegalArgumentException("improper character for SafeId: $ch")
                        codeTable[ch.code].also {
                            if (it < 0) throw IllegalArgumentException("improper character for SafeId: $ch")
                        }
                    }
                }
                buffer = buffer or (i shl bits)
                bits += 5
                if (bits >= 8) {
                    result.add((buffer and 0xFF).toByte())
                    bits -= 8
                    buffer = buffer shr 8
                }
            }
            return result.toByteArray()
        }

        /**
         * Decode and represent as Long value
         */
        fun toLong(encoded: String): Long {
            var data = decode(encoded)
            if (data.size < 8) {
                val fill = ByteArray(8 - data.size)
                fill.fill(0)
                data = fill + data
            }
            return bytesToLong(data)
        }

        /**
         * Decode and represent as Int value
         */
        fun toInt(encoded: String): Int = toLong(encoded).toInt()
    }
}

/**
 * Encode to Long using big-endian notation. Note that [SafeId.toLong] will properly work despite [trimLeadingZeroes]
 * value.
 * @param trimLeadingZeroes do not encode not significant leading zeroes.
 */
fun Long.toSafeId(trimLeadingZeroes: Boolean = true): String {
    var data = longToBytes(this)
    val i = data.indexOfFirst { it != 0.toByte() }
    if (trimLeadingZeroes) {
        when {
            i >= 0 -> data = data.sliceArray(i..7)
            i < 0 -> data = byteArrayOf(0)
        }
    }
    return SafeId[data]
}

/**
 * Encode to Int using big-endian notation. Note that [SafeId.toInt] will properly work despite [trimLeadingZeroes]
 * value.
 * @param trimLeadingZeroes do not encode not significant leading zeroes.
 */
fun Int.toSafeId(trimLeadingZeroes: Boolean = true): String {
    var data = longToBytes(this.toLong()).sliceArray(4..7)
    val i = data.indexOfFirst { it != 0.toByte() }
    if (trimLeadingZeroes) {
        when {
            i >= 0 -> data = data.sliceArray(i..3)
            i < 0 -> data = byteArrayOf(0)
        }
    }
    return SafeId[data]
}
