@file:OptIn(DelicateCoroutinesApi::class)
@file:Suppress("EXPERIMENTAL_API_USAGE")

package net.sergeych.parsec

import kotlinx.coroutines.*
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.datetime.Instant
import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import net.sergeych.boss_serialization.BossDecoder
import net.sergeych.boss_serialization_mp.*
import net.sergeych.mp_logger.*
import net.sergeych.unikrypto.*
import kotlin.random.Random

/**
 * Interface to key address provider (depending on the parsec version, it could be pre-shared
 * addresses, pre-shared service contract origin in which case the service provides actual contract,
 * and caller checks it and retrieves addresses, or UNS2 name, that allowa caller to get
 * contract directly from the universa network.
 */
interface ServiceIdProvider {
    /**
     * Get known valid service key addresses. Could be cached.
     */
    val keyIds: List<KeyIdentity>

    /**
     * Session wants provider to refresh known addresses, e.g. re-obtain contracts, keys, whatever
     * depending on the parsec version. Default version does nothing.
     */
    suspend fun refresh() {
    }
}

@Serializable
class RequestSckResult(val context: ByteArray, val POWTask: POWTask)

@Suppress("unused")
@Serializable
class RequestSckArgs(val SCKAddress: ByteArray, val testMode: Boolean = false)

@Serializable
@Suppress("unused")
class RegisterSCKArgs(val context: ByteArray, val signedRecord: ByteArray)

@Serializable
class RegisterSCKResult(val sessionId: String)

@Serializable
data class CreateTSKResult(val clientNonce: ByteArray,val encryptedResult: ByteArray)

@Serializable
data class TSKResultPayload(
    @SerialName("TSKExpiresAt")
    val sessionExpiresAt: Instant,
    val TSK: ByteArray)

/**
 * The parsec 1.* session. The session is the same protocol, except for way to obtain valid service
 * key address that varies, and implemented outside this class as a given [ServiceIdProvider]
 * that sholuld ]
 */
class ParsecSession(builderCallback: Builder.() -> Unit) : ParsecConnection, Loggable by LogTag("PARS") {

    class Builder {
        var storage: KVBinaryStorage = MemoryKVBinaryStorage()
        var testMode = false
        var keyStrength = 4096
        var url: String? = null
        var rootConnection: RootConnection? = null
        var idProvider: ServiceIdProvider? = null
        var serviceIds: List<KeyIdentity>? = null
    }

    private val builder = Builder().also { it.builderCallback() }
    internal val rootConnection: RootConnection
    private val serviceIds: ServiceIdProvider

    private val storage = KVStorage(builder.storage)

    var TSK: ByteArray? by optStored(storage, ".p1.TSK")
        private set

    val TSKOrException: ByteArray get() = TSK ?: throw IllegalStateException("TSK is not negotiated")

    private var SCK: PrivateKey? by optStored(storage, ".p1.SCK")

    private var sessionId: String? by stored(storage, ".p1.SID")

    var sckGenerationsCount: Int = 0
        private set

    var tskGenerationCount: Int = 0
        private set

    var sessionExpiresAt: Instant? = null
        private set

    val connectedState = MutableStateFlow(false)

    private var sessionEndpoint = CompletableDeferred<Endpoint>()

    private val keyStrength by builder::keyStrength
    private val testMode by builder::testMode

    init {
        rootConnection = builder.url?.let {
            if (builder.rootConnection != null)
                throw IllegalArgumentException("allowed either url or rootConnection, not both")
            RootConnection(it)
        } ?: run {
            builder.rootConnection
                ?: throw IllegalArgumentException("parsec session builder MUST provide either rootConnection or url")
        }

        serviceIds = builder.idProvider ?: builder.serviceIds?.let {
            object : ServiceIdProvider {
                override val keyIds: List<KeyIdentity>
                    get() = it
            }
        } ?: throw IllegalArgumentException("required either idProvider or keyIds")


        GlobalScope.launch {
            try {
                sessionEndpoint.complete(connect())
            } catch (e: Exception) {
                exception { "Fatal exception in parsec connect(), parsec protocol aborted" to e }
            }
        }
    }

    private suspend fun connect(): Endpoint {
        debug { "starting parsec connect" }
        while (true) {
            debug { "connect loop iteration started" }
            try {
                TSK?.let { tsk ->
                    debug { "use TSK" }
                    sessionId?.let { sid ->
                        val ep = Endpoint(rootConnection, sessionKey = tsk, authToken = sid)
                        try {
                            val result = ep.call("getSessionInfo")
                            debug { "session established (${sessionId}): info: $result" }
                            this.sessionExpiresAt = result["expiresAt"] as? Instant
                            // success: we got an endpoint!
                            connectedState.value = true
                            return ep
                        } catch (e: ParsecRemoteException) {
                            debug { "session can't be connected: $e, dropping TSK" }
                        } catch (e: Exception) {
                            debug { "network error while establishing session" }
                            throw e
                        }
                    } ?: debug { "inconsistent TKS: no sessionId, dropping" }
                    // if we get there, we have to clear TSK
                    debug { "dropping TSK" }
                    TSK = null
                }
                // NO TSK
                debug { "NO TSK" }
                return startSession()
            } catch (e: Exception) {
                e.printStackTrace()
                debug { "failed to establish the connection: $e" }
                delay(5000)
            }
        }
    }

    suspend fun startSession(): Endpoint {
        if (SCK != null) {
            connectWithSCK()?.let { return it }
            debug { "session SCK is invalid (reconnection failed), dropping it" }
            SCK = null
        }
        return generateNewSCK()
    }

    suspend fun connectWithSCK(): Endpoint? {
        try {
            debug { "creating TSK for $sessionId" }
            val clientNonce = Random.nextBytes(32)
            val result = rootConnection.call(
                "createTSK",
                "signedRecord" to SignedRecord.pack(
                    SCK ?: fatal("SCK must be set by now"),
                    "clientNonce" to clientNonce,
                    "sessionId" to (sessionId ?: fatal("sression ID should exist")),
                    "knownServiceKeys" to serviceIds.keyIds.map { it.id },
                )
            )
            val signedResult = SignedRecord.unpack(
                result.getByteArray("signedEncryptedResult") ?: fatal("invalid createTSK answer")
            )
            val receivedNonce = signedResult.payload.getByteArray("clientNonce")
            if (!(clientNonce contentEquals receivedNonce))
                throw ParsecAuthenticationException("invalid nonce")

            tskGenerationCount++
            for (a in serviceIds.keyIds) {
                if (signedResult.publicKey.id == a) {
                    val payload = signedResult.decode<CreateTSKResult>()
                    debug { "found a valid service key address, connection is approved, SCK: $SCK" }
                    val plainResult = SCK?.etaDecrypt(payload.encryptedResult)?.decodeBoss<TSKResultPayload>()
                        ?: fatal("SCK must be set")
                    TSK = plainResult.TSK
                    sessionExpiresAt = plainResult.sessionExpiresAt
                    debug { "got a TSK, restarting session for it. Expiration (ignored) is $sessionExpiresAt" }
                    // now we have SCK and TSK, lets run another attempt
                    return connect()
                }
            }
            if (tskGenerationCount < 3) {
                debug { "tsk generation failed. attempt #$tskGenerationCount. Will retry" }
                // this may mean that our possibily cached serviceIds information is outdated:
                serviceIds.refresh()
            } else
                throw ParsecAuthenticationException("no suitable service key found")

        } catch (e: ParsecRemoteException) {
            debug { "service rejected our SCK->TSK request: ${e.message}, resetting SCK" }
            SCK = null
            // generate new key - it will also establish new session.
            // stack should not get deep as it will either succeed or throw exception
            generateNewSCK()
        }
        return null
    }

    /**
     * Generate new SCK (e.g. new session, new TSK, etc) so it always create endpoint or
     * throws IOException on the network failure.
     */
    private suspend fun generateNewSCK(): Endpoint {
        //
        SCK = null
        while (true) {
            // we loop until our key will not ba accepted
            val sck = AsymmetricKeys.generate(keyStrength)
            try {
                val test: BossStruct = rootConnection
                    .exec("requestSCK", RequestSckArgs(sck.publicKey.id.id, testMode))
//                println("!!!? ${test.getStruct("POWTask")}")
//                println("!!!? ${test["context"]}")
                val rsr = BossDecoder.decodeFrom<RequestSckResult>(test)
//                val rsr: RequestSckResult = rootConnection
//                    .exec("requestSCK", RequestSckArgs(sck.publicKey.id.id, testMode))
                val powTask = rsr.POWTask
                debug { "Got a POWTask, calculating solution length=${powTask.length}" }
                val solutionRecord = SignedRecord.pack(sck, "POWResult" to powTask.solve())
                debug { "solution calculated" }
                val result: RegisterSCKResult =
                    rootConnection.exec("registerSCK", RegisterSCKArgs(rsr.context, solutionRecord))
                sckGenerationsCount++
                SCK = sck
                sessionId = result.sessionId
                return startSession()
            } catch (e: ParsecRemoteException) {
                warning { "remote has rejected our key! ($e)" }
            }
        }
    }

    override suspend fun call(command: String, params: BossStruct): BossStruct {
        try {
            return sessionEndpoint.await().call(command, params)
        } catch (e: ParsecRemoteException) {
            warning { "Service error: $e" }
            when (e.code) {
                "parsec_missing_tsk", "bad_sck_key", "session_not_found" -> {
                    warning { "session lost with code ${e.code}, attempting to reestablish" }
                    TSK = null
                    // protocol is now incapable of performing any commands and requires re-establishing connection, so:
                    reconnect()
                    // and repeat current call
                    return call(command, params)
                }
                else -> throw e
            }
        }
    }

    @Suppress("unused")
    suspend fun waitReady() {
        // session is ready when/if its endpoint is ready
        sessionEndpoint.await()
    }

    fun onConnected(block: ()->Unit) {
        sessionEndpoint.invokeOnCompletion { block() }
    }

    suspend fun reconnect() {
        // - make all competing requests queue until connected
        sessionEndpoint = CompletableDeferred()
        // - reconnect it
        sessionEndpoint.complete(connect())
    }

    suspend fun clearTSK() {
        TSK = null
        reconnect()
    }

    @Suppress("unused")
    val currentSessionId: String?
        get() = sessionId

    @Suppress("unused")
    val sessionKey: SymmetricKey
        get() = if (sessionEndpoint.isCompleted) sessionEndpoint.getCompleted().key
        else throw IllegalStateException("session key not set")

    companion object {

//        @Suppress("unused")
//        fun badFormat(text: String = "invalid data format"): Nothing {
//            throw ParsecException("parsec_invalid_format", text)
//        }

        fun fatal(message: String): Nothing {
            throw ParsecException("parsec_protocol_error", "Protocol error: $message")
        }
    }
}