@file:Suppress("unused")
@file:OptIn(DelicateCoroutinesApi::class)

package net.sergeych.ugdx

import kotlinx.coroutines.*
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.asStateFlow
import net.sergeych.boss_serialization_mp.BossStruct
import net.sergeych.boss_serialization_mp.KVStorage
import net.sergeych.mp_logger.*
import net.sergeych.mptools.CachedExpression
import net.sergeych.mptools.reportExceptions
import net.sergeych.parsec.ParsecSession
import net.sergeych.parsec.required
import net.sergeych.ugdx.client.PasswordKeys
import net.sergeych.ugdx.shared.*
import net.sergeych.ugdx.shared.UgdxCommand.*
import net.sergeych.ugdx.shared.api.OrderIdArgs
import net.sergeych.ugdx.shared.api.cmdGetOrder
import net.sergeych.unikrypto.AheadPrivateKeyGenerator
import net.sergeych.unikrypto.Keyring
import net.sergeych.unikrypto.SymmetricKeys

/*
 * Copyright (C) 2019-2022 by Sergey S. Chernov (real.sergeych@gmail.com), all rights reserved.
 *
 * Class 2 myonly.cloud registration/login dance.
 *
 * This implementation is exclusively licensed to be used in UnitedAssets project only.
 *
 * Note to developers:
 *
 * This API is consumed by various ugdx client software. The network-exposed part should not be in this file.
 */

@Suppress("unused")
sealed class RegistrationResult {
    object LoginInvalid : net.sergeych.ugdx.RegistrationResult()
    object LoginUnavailable : net.sergeych.ugdx.RegistrationResult()
    object InvalidInviteCode : net.sergeych.ugdx.RegistrationResult()
    class Error(val text: String, val cause: ApiException? = null) : net.sergeych.ugdx.RegistrationResult()
    class OK(val account: net.sergeych.ugdx.UserAccount) : net.sergeych.ugdx.RegistrationResult()
}

sealed class SignInResult {
    object BadCredentials : net.sergeych.ugdx.SignInResult()
    class Error(val text: String) : net.sergeych.ugdx.SignInResult()
    class OK(val account: net.sergeych.ugdx.UserAccount) : net.sergeych.ugdx.SignInResult()
}

var useLocalServer = true
var localServerUrl = "http://localhost:8080/api/p1"

object GuestAPI : LogTag("GtAPI") {

    fun setup(initializer: ParsecSession.Builder.() -> Unit) {

        val s = ParsecSession {
            initializer()
            GuestAPI.storage = KVStorage(storage)
        }
        s.onConnected {
            GuestAPI.connection.complete(s)
        }
        GuestAPI.savedUserAccountInfo?.let { uai ->
            info { "found saved user account: ${uai.loginName}, remembering user" }
            GlobalScope.launch { reportExceptions {
                net.sergeych.ugdx.GuestAPI.updateAccount(
                    net.sergeych.ugdx.UserAccount(
                        uai
                    )
                )
            } }
        }
            ?: run { net.sergeych.ugdx.GuestAPI.states.value = net.sergeych.ugdx.ServiceState.NotSignedIn }
    }

    val connection = CompletableDeferred<ParsecSession>()
    var storage: KVStorage? = null
        private set

    var savedUserAccountInfo: AccountInfo?
        get() = net.sergeych.ugdx.GuestAPI.storage?.get<AccountInfo>("savedUserAccount")
        set(value) {
            net.sergeych.ugdx.GuestAPI.storage?.set("savedUserAccount", value)
        }

    val keyGenerator = AheadPrivateKeyGenerator(4096, 2)

    /**
     * Here API code posts servce state changes:
     */
    internal val states = MutableStateFlow<net.sergeych.ugdx.ServiceState>(net.sergeych.ugdx.ServiceState.Loading)

    /**
     * The flow ot state changes, see [ServiceState]. It it possible to check current service state via
     * `StateFlow.state` or to collect state changes:
     * ~~~
     * stateFlow.collect { newState ->
     *   println("Service not in state: $newState")
     * }
     * ~~~
     * note that state flow guarantees that collects immediately gets the current state, so _there is no need to check_
     * `flowState.state` if you do collect the flow, you get it anyway.
     */
    val stateFlow: StateFlow<net.sergeych.ugdx.ServiceState> = net.sergeych.ugdx.GuestAPI.states.asStateFlow()


    /**
     * Check that some login name is available. No guarantees that it will be available on the next call ;)
     * Server tries to reduce "look-alike" login name collisions, so many logins could not be available
     * due to it.
     */
    suspend fun isLoginNameAvailable(login: String): Boolean =
        CHECK_NAME_AVAILABLE.call<BossStruct>(net.sergeych.ugdx.GuestAPI.connection, "loginName" to login)
            .required("loginNameAvailable")

    suspend fun isValidInviteCode(code: String): Boolean =
        CHECK_INVITE_CODE.call<BossStruct>(net.sergeych.ugdx.GuestAPI.connection, "inviteCode" to code)
            .required("isValid")

    /**
     * start background generation of keys that are used in the registration or wherever else
     * strong private keys are used.
     */
    fun startGeneratingKeys() {
        net.sergeych.ugdx.GuestAPI.keyGenerator.start()
    }

    /**
     * Perform registration for a given login and password. This is a very long operation as it may include
     * wait time for parsec connection to establish, generating sign in private key (strong) and several
     * network calls.
     */
    suspend fun register(
        login: String, password: String, inviteCode: String? = null,
        rounds: Int = 75000, testMode: Boolean = false,
    ): net.sergeych.ugdx.RegistrationResult {
        return coroutineScope {
            // In parrallel (potentially):
            // - generate new key
            // if it is not yet started, start it now
            net.sergeych.ugdx.GuestAPI.keyGenerator.start()
//            val deferredLoginKey = async { keyGenerator.getKey() }
//            val deferredMainKey = async { keyGenerator.getKey() }
            // - Create 3 secret keys
            val derivationParams = net.sergeych.ugdx.DerivationParams.Companion.random(if (testMode) 500 else rounds)
            val deferredPasswordKeys = async { PasswordKeys.deriveFrom(password, derivationParams) }

            // when everything is ready, we can request registration:
            val passwordKeys = deferredPasswordKeys.await()
            // generator was working in background already
            val mainKey = net.sergeych.ugdx.GuestAPI.keyGenerator.getKey()
            val loginKey = net.sergeych.ugdx.GuestAPI.keyGenerator.getKey()

            val kr = Keyring("main" to mainKey, "data" to SymmetricKeys.random())
            try {
                val encryptedKeyring = passwordKeys.encryptKeyring(kr)
                val result: AccountInfo = AUTH_REGISTER.call(
                    net.sergeych.ugdx.GuestAPI.connection, SignedRecordArgs(
                        loginKey,
                        RegisterArgsPayload(
                            login,
                            passwordKeys.token,
                            passwordKeys.encryptLoginObject(loginKey),
                            encryptedKeyring,
                            mainKey.publicKey,
                            derivationParams,
                            inviteCode,
                            testMode
                        ),
                        net.sergeych.ugdx.GuestAPI.connection.await().TSKOrException
                    )
                )
                net.sergeych.ugdx.RegistrationResult.OK(
                    net.sergeych.ugdx.UserAccount(result, passwordKeys)
                        .also { net.sergeych.ugdx.GuestAPI.updateAccount(it) }
                )
            } catch (x: ApiException) {
                when (x.code) {
                    ErrorCode.LOGIN_NAME_UNAVAILABLE -> net.sergeych.ugdx.RegistrationResult.LoginUnavailable
                    ErrorCode.INVALID_INVITE_CODE -> net.sergeych.ugdx.RegistrationResult.InvalidInviteCode
                    else -> net.sergeych.ugdx.RegistrationResult.Error(x.text, x)
                }
            } catch (x: Exception) {
                x.printStackTrace()
                net.sergeych.ugdx.RegistrationResult.Error("unexpected exception: $x")
            }
        }
    }

    /**
     * Log in to usass environment. The process could take considerable time as it _may include_ parsec establishing
     * and requires sign in key decryption and login proof signing (slow).
     */
    suspend fun signIn(login: String, password: String): net.sergeych.ugdx.SignInResult =
        net.sergeych.ugdx.GuestAPI.signIn(login, password, 0)

    suspend private fun signIn(login: String, password: String, tryNumber: Int): net.sergeych.ugdx.SignInResult = try {
        val session = net.sergeych.ugdx.GuestAPI.connection.await()
        // Login dance is rather complex. First, we request KDF parameters to properly derive keys and tokens:
        val signInRequestResult: SignInRequestResult =
            REQUEST_SIGN_IN.call(net.sergeych.ugdx.GuestAPI.connection, SignInRequestArgs(login))
        val passwordKeys = signInRequestResult.kdfParams.deriveFrom(password)
        // now we have token to het login object and key to decrypt is, requesting login object:
        val signInObjectResult: SignInObjectRequestResult =
            REQUEST_SING_IN_OBJECT.call(
                net.sergeych.ugdx.GuestAPI.connection, SignInObjectRequestArgs(
                    login,
                    passwordKeys.seedToken(signInRequestResult.tokenSeed),
                )
            )
        // we got a login object, now we can decrypt it and sign in the singIn request ;)
        val singInKey = passwordKeys.decryptLoginObject(signInObjectResult.encryptedLoginObject).privateKey
        // and only new we can sign in!
        val accountInfo: AccountInfo = PERFORM_SIGN_IN.call(
            net.sergeych.ugdx.GuestAPI.connection,
            SignedRecordArgs(singInKey, SignInPayload(null), session.TSKOrException)
        )
        // Hurray. We're in. publish our main key if need
        net.sergeych.ugdx.SignInResult.OK(
            net.sergeych.ugdx.UserAccount(accountInfo, passwordKeys).also { acc ->
                net.sergeych.ugdx.GuestAPI.updateAccount(acc)
                // Temp: fix account keys, new registration does it already, so later
                // TODO: remove when all accounts are upgraded
                acc.getParty(acc.partyId)?.let { me ->
                    if (me.optPublicKey == null || me.optPublicKey.id != acc.mainKey().id) {
                        info { "Correcting main key on the server:" }
                        val mk = acc.mainKey()
                        net.sergeych.ugdx.cmdUpdateMainKey(mk.publicKey)
                        info { "Main key is published: ${mk.publicKey.id}" }
                    }
                }
                    ?: warning { "failed to get 'me' party, pls report error to developers, userId=${acc.userId}, partyId=${acc.partyId}" }
            }
        )
    } catch (x: ApiException) {
        when (x.code) {
            ErrorCode.INVALID_CREDENTIALS, ErrorCode.NOT_FOUND -> net.sergeych.ugdx.SignInResult.BadCredentials
            ErrorCode.RETRY_LOGIN -> {
                /**
                 * It _maight_ happen that server side seed became ourdated. Though we are trying to avoid it,
                 * so we try once more
                 */
                if (tryNumber > 1)
                    net.sergeych.ugdx.SignInResult.BadCredentials
                else
                    net.sergeych.ugdx.GuestAPI.signIn(login, password, tryNumber + 1)
            }
            else -> {
                error { "Unexpected API error: $x" }
                net.sergeych.ugdx.SignInResult.Error("unexpected api error: $x")
            }
        }
    } catch (x: Exception) {
        exception { "unexpected error while loggin in" to x }
        net.sergeych.ugdx.SignInResult.Error("unexpected error: $x")
    }

    /**
     * This function just checks the connection is still maintained and update shared status as need. Client
     * should not call it unless it wants to re-check it immediately. The result is the same what `stateFlow.value`
     * returns.
     */
    suspend fun refreshStatus(): net.sergeych.ugdx.ServiceState {
        val status = CHECK_STATUS.call<CheckStatusResult>(net.sergeych.ugdx.GuestAPI.connection)
        info { "Status checked: $status" }
        if (net.sergeych.ugdx.GuestAPI.currentAccount != null && status.accountInfo != null &&
            status.accountInfo.id != net.sergeych.ugdx.GuestAPI.currentAccount?.accountInfo?.id
        ) {
            warning { "logged in account differs from current, force sign out" }
            net.sergeych.ugdx.GuestAPI.signOut()
        } else {
            if (net.sergeych.ugdx.GuestAPI.currentAccount == null && status.accountInfo != null) {
                // logging in without proper procedure: not good
                warning { "connection status is logged in while we didn't do it, logging out" }
                net.sergeych.ugdx.GuestAPI.signOut()
            } else if (net.sergeych.ugdx.GuestAPI.currentAccount != null && status.accountInfo == null) {
                info { "we are logged out by some other instance" }
                net.sergeych.ugdx.GuestAPI.signOut()
            }
        }
        return net.sergeych.ugdx.GuestAPI.stateFlow.value
    }

    var currentAccount: net.sergeych.ugdx.UserAccount? = null
        private set

    val account: net.sergeych.ugdx.UserAccount
        get() = net.sergeych.ugdx.GuestAPI.currentAccount ?: ErrorCode.NOT_SIGNED_IN.raise()


    val userAccountOrThrow: net.sergeych.ugdx.UserAccount get() = net.sergeych.ugdx.GuestAPI.currentAccount ?: ErrorCode.NOT_SIGNED_IN.raise()

    private fun updateAccount(newAccount: net.sergeych.ugdx.UserAccount?) {
        if (net.sergeych.ugdx.GuestAPI.currentAccount != newAccount) {
            if (newAccount != null) {
                // turning to signed in
                if (net.sergeych.ugdx.GuestAPI.currentAccount != null) {
                    println("clearing user data on change acc")
                    net.sergeych.ugdx.GuestAPI.clearUserData()
                    net.sergeych.ugdx.GuestAPI.currentAccount = null
                }
                net.sergeych.ugdx.GuestAPI.currentAccount = newAccount
                net.sergeych.ugdx.GuestAPI.savedUserAccountInfo = newAccount.accountInfo
                net.sergeych.ugdx.GuestAPI.states.value = net.sergeych.ugdx.ServiceState.SignedIn(newAccount)
            } else {
                // changing to signed out
                println("clearing user data as requested")
                net.sergeych.ugdx.GuestAPI.currentAccount = null
                net.sergeych.ugdx.GuestAPI.clearUserData()
                net.sergeych.ugdx.GuestAPI.states.value = net.sergeych.ugdx.ServiceState.NotSignedIn
            }
        }
    }

    private fun clearUserData() {
        net.sergeych.ugdx.GuestAPI.savedUserAccountInfo = null
    }

    suspend fun signOut() {
        SIGN_OUT.call<BossStruct>(net.sergeych.ugdx.GuestAPI.connection)
        net.sergeych.ugdx.GuestAPI.connection.await().clearTSK()
        net.sergeych.ugdx.GuestAPI.updateAccount(null)
    }

    private val currencyCache = CachedExpression<GetCurrenciesResult>()

    private suspend fun getCurrencyData(): GetCurrenciesResult = net.sergeych.ugdx.GuestAPI.currencyCache.get {
        GET_CURRENCIES.call(net.sergeych.ugdx.GuestAPI.connection)
    }

    /**
     * Request list of supported currencies. This list does not change often, and could be locally cached
     */
    suspend fun getCurrencies(): List<Currency> = net.sergeych.ugdx.GuestAPI.getCurrencyData().currencies

    /**
     * Get a currency that is used as a base to quote in [Currency.baseRate].
     */
    suspend fun getBaseCurrency(): Currency = net.sergeych.ugdx.GuestAPI.getCurrencyData().baseCurrency

    /**
     * Get the currency by code or throw `NoSuchElementException`
     */
    suspend fun currencyWithCode(code: String) = net.sergeych.ugdx.GuestAPI.getCurrencies().first { it.code == code }

    /**
     * Delete test user. If there is a user but it was not created in test mode, will not delete it and will throw
     * exception instead. It is ok to try to delete unexesting user, but it might not work on non-test sites.
     * @return true if the user was ectually deleted
     */
    suspend fun deleteTestUser(loginName: String): Boolean =
        DELETE_TEST_USER.call<BossStruct>(net.sergeych.ugdx.GuestAPI.connection, "loginName" to loginName)
            .required<Boolean>("result")

    /**
     * If there is no logged in user, it will fail, so client should call public [UserAccount] method instead
     */
    internal suspend fun getOrder(orderId: Long): OrderRecord = cmdGetOrder(OrderIdArgs(orderId))
}