package net.sergeych.ugdx

import net.sergeych.ugdx.shared.EmptyResult
import net.sergeych.ugdx.shared.ErrorCode
import net.sergeych.ugdx.shared.api.ApiCommand
import kotlinx.datetime.Instant
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.mptools.CachedExpression
import net.sergeych.unikrypto.KeyIdentity
import net.sergeych.unikrypto.PublicKey

/**
 * Party is a "public" representatino of a client that is not connected to its loginName or any searchable id.
 * It is used to disclose no more and no less information to identify participants of invoices, transfers, etc.
 * @param id partyId, some random and unique immutable string assigned bu the service.
 * @param name the most recent name of the party (varies, the client can change its name)
 * @param publicKeyId the id of the public key this party signs documents, also can be used to encrypt private messages
 *                    to.
 * @param optPublicKey internal usage only, pleas use [publicKey] instead
 *
 */
@Serializable
data class Party(
    val id: String,
    val name: String?,
    val publicKeyId: KeyIdentity,
    val optPublicKey: PublicKey? = null,
) {
    @Transient
    private var _publicKey = CachedExpression(initialValue = optPublicKey)

    /**
     * Get the public key, for example, to check signature or encrypt a private message to the party. Requests and
     * caches a key if need.
     */
    suspend fun publicKey(): PublicKey? =
        _publicKey.optGet {
            optPublicKey ?: cmdPartyGetKey(PartyIdArgs(id)).publicKey
        }

    /**
     * Return most recent data about the party.
     */
    suspend fun refresh(): Party {
        return get(id, _publicKey.cachedOrNull() != null )
    }

    fun overridePublicKey(publicKey: PublicKey) {
        if( publicKey.id != publicKeyId ) ErrorCode.INVALID_PARAMETER.raise("public key does not match party")
        _publicKey = CachedExpression(initialValue = publicKey)
    }

    override fun equals(other: Any?): Boolean {
        return if( other is Party)
            return other.id == id
        else false
    }

    override fun hashCode(): Int {
        return id.hashCode()
    }

    companion object {
        suspend fun get(id: String, includeKey: Boolean = false): Party = cmdPartyGet(PartyIdArgs(id, includeKey))
    }
}

@Serializable
data class PartyIdArgs(val partyId: String, val extended: Boolean = false)

@Serializable
data class PublicKeyRecord(val keyId: KeyIdentity,val publicKey: PublicKey?,val expiresAt: Instant?)

val cmdPartyGetKey = ApiCommand<PartyIdArgs, PublicKeyRecord>("Party.getKey")

val cmdPartyGet = ApiCommand<PartyIdArgs, Party>("Party.get")
val cmdPartyGetByLogin = ApiCommand<PartyIdArgs, Party>("Party.getByLogin")

val cmdUpdateMainKey = ApiCommand<PublicKey, EmptyResult>("User.updateMainKey")