@file:OptIn(ExperimentalSerializationApi::class)

package net.sergeych.ugdx.client

import net.sergeych.ugdx.DerivationParams
import kotlinx.datetime.Clock
import kotlinx.datetime.Instant
import kotlinx.serialization.ExperimentalSerializationApi
import kotlinx.serialization.Serializable
import net.sergeych.boss_serialization_mp.BossEncoder
import net.sergeych.boss_serialization_mp.decodeBoss
import net.sergeych.unikrypto.*
import kotlin.math.sign

/**
 * The set of keys that is generated from the password and utililty functions that helps to use proper
 * keys in different scenarios.
 */
data class PasswordKeys(

    /**
     * Key used to request sign in object from the service. Do not use it direcrly, use [token] and [seedToken]
     * instead.
     */
    val signInAccessKey: SymmetricKey,
    /**
     * Key to decrypt sign in object; do not use it directly, use [encryptLoginObject] and [decryptLoginObject] for
     * guaranteed compatibility
     */
    val signInObjectDecryptionKey: SymmetricKey,
    /**
     * Key to access encrypted user data stored in the cloud (e.g. his main key, etc). This key should be used to
     * store and restore objects stored in the internal ugdx cloud, primarily, keys to sign documents, contracts,
     * transfers, etc. In particular, [encryptKeyring] and [decryptKeyring] use this key.
     */
    val dataAccessKey: SymmetricKey
) {

    @Serializable
    data class LoginObject(
        val privateKey: PrivateKey,
        val generatedAt: Instant
    )

    /**
     * The token used to open access to encrypted login object: with this token (seeded, see [seedToken])
     * the service will provide encrypted login object that could be then decrypted to perform actual login.
     *
     * NTW. Token DOES NOT weaken strength of the signInAccessKey so it could be used to independently
     * encrypt some future information.
     */
    val token by lazy { signInAccessKey.id.id }

    /**
     * For a true paranoyal security we transform login token by "seeding" it with some random data from
     * a server side. This prevent a class of repeating data attacks from a partially compsomised client
     * e.g. chrome browser ;)
     */
    fun seedToken(seed: ByteArray): ByteArray = net.sergeych.ugdx.shared.seedToken(seed, token)

    fun encryptLoginObject(signInKey: PrivateKey): ByteArray =
        signInObjectDecryptionKey.etaEncrypt(BossEncoder.encode(LoginObject(signInKey, Clock.System.now())))

    fun decryptLoginObject(encryptedLoginObject: ByteArray): LoginObject =
        signInObjectDecryptionKey.etaDecrypt(encryptedLoginObject).decodeBoss()

    fun encryptKeyring(kr: Keyring): ByteArray = dataAccessKey.etaEncrypt(BossEncoder.encode(kr))

    fun decryptKeyring(encrypted: ByteArray): Keyring = dataAccessKey.etaDecrypt(encrypted).decodeBoss()

    companion object {
        /**
         * Generate keys triple from the password and stored derivation params
         */
        suspend fun deriveFrom(password: String, params: DerivationParams = DerivationParams.random()): PasswordKeys {
            val (k1, k2, k3) = Passwords.deriveKeys(password, 3, params.rounds, params.hashAlgorithm, params.salt)
            return PasswordKeys(k1, k2, k3)
        }
    }
}