package net.sergeych.ugdx.realtime

import io.ktor.client.features.websocket.*
import io.ktor.http.cio.websocket.*
import kotlinx.coroutines.CancellationException
import kotlinx.coroutines.CompletableDeferred
import kotlinx.coroutines.delay
import kotlinx.coroutines.flow.*
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock
import kotlinx.serialization.Serializable
import net.sergeych.boss_serialization.BossDecoder
import net.sergeych.boss_serialization_mp.BossEncoder
import net.sergeych.boss_serialization_mp.BossStruct
import net.sergeych.boss_serialization_mp.decodeBoss
import net.sergeych.mp_logger.*
import net.sergeych.mp_tools.globalLaunch
import net.sergeych.mptools.toDump
import net.sergeych.parsec.ParsecSession
import net.sergeych.ugdx.shared.ErrorCode
import net.sergeych.ugdx.shared.api.AsyncBlock
import net.sergeych.ugdx.shared.api.PayloadBlock
import net.sergeych.ugdx.shared.api.PushPayload
import net.sergeych.ugdx.shared.trader.CurrencyPair
import net.sergeych.ugdx.shared.trader.StatSampleRate
import net.sergeych.unikrypto.SymmetricKey
import net.sergeych.unikrypto.SymmetricKeys

@Serializable
data class WatchTraderArgs(val currencyPair: CurrencyPair, val on: Boolean)

@Serializable
data class WatchTraderLastStickArgs(val currencyPair: CurrencyPair, val statSampleRate: StatSampleRate)

internal class RealtimeChannel private constructor(
    private val socket: DefaultClientWebSocketSession,
    private val parsecSession: ParsecSession,
    oldChannel: RealtimeChannel? = null,
) : LogTag("RTCHN") {

    private var lastPackedTsk: ByteArray? = null
    private var lastTsk: SymmetricKey? = null

    val tsk: SymmetricKey
        get() {
            val packed = parsecSession.TSKOrException
            var key = if (lastPackedTsk != null && packed contentEquals lastPackedTsk) lastTsk else null
            if (key == null) {
                key = SymmetricKeys.create(packed)
                lastTsk = key
                lastPackedTsk = packed
            }
            return key
        }

    private val access = Mutex()
    private var lastId = 0
    private val deferred = HashMap<Int, CompletableDeferred<ByteArray?>>()

    private val commandsReady = CompletableDeferred<Boolean>()

    suspend fun rawCommand(cmd: String, args: BossStruct = BossStruct.EMPTY): ByteArray? {
        val crresult = commandsReady.await()
        debug { "commandsready status: ${commandsReady.isActive} ${commandsReady.isCompleted} ${commandsReady.isCancelled}"}
        debug { "commandsreadt result $crresult" }
        return access.withLock {
            val id = lastId++
            val result = CompletableDeferred<ByteArray?>()
            deferred[id] = result
            sendPayload(PayloadBlock.Request(id, cmd, args))
            result
        }.await()
    }

    suspend fun command(cmd: String, vararg args: Pair<String, Any>): ByteArray? =
        rawCommand(cmd, BossStruct.from(*args))

    suspend inline fun <reified A : Any> command(cmd: String, args: A): ByteArray? =
        rawCommand(cmd, BossEncoder.encodeToStruct(args))

    private suspend fun sendPayload(payloadBlock: PayloadBlock) {
        sendBlock(
            AsyncBlock.Data(tsk.etaEncrypt(BossEncoder.encode(payloadBlock)))
        )
    }

    private suspend fun sendAck() {
        sendBlock(AsyncBlock.Ack(
            parsecSession.currentSessionId ?: ErrorCode.INTERNAL_ERROR.raise("parsec session has no ID"),
            tsk.etaEncrypt("Session: ${parsecSession.currentSessionId}")
        ))
    }

    private suspend fun start(onConnectionStateChanged: suspend (RealtimeChannel?) -> Unit) {
        debug { "websocket connection obtained, handshaking: $socket" }
        for (frame in socket.incoming) {
            if (frame is Frame.Binary) {
                processBlock(BossDecoder.decodeFrom(frame.data), onConnectionStateChanged)
            } else {
                warning { "ignoring wrong frame type $frame" }
            }
        }
        debug { "closing websocket" }
        onConnectionStateChanged(null)
    }


    private suspend fun processBlock(block: AsyncBlock, onConnectionStateChanged: suspend (RealtimeChannel?) -> Unit) {
        when (block) {
            is AsyncBlock.Nack -> sendAck()
            is AsyncBlock.Ack -> sendAck()
            is AsyncBlock.Accept -> {
                debug { "PARSEC WS connection established" }
                onConnectionStateChanged(this)
                debug { "setting commandsReady" }
                commandsReady.complete(true)
            }
            is AsyncBlock.Data -> {
                val decrypted = kotlin.runCatching { tsk.etaDecrypt(block.encrypted) }
                if (decrypted.isFailure) {
                    warning { "failed to decrypt data block, the channel might need be reestablished" }
                    socket.close()
                } else {
                    val decryptedData = decrypted.getOrThrow()
                    val payloadBlock = runCatching { decryptedData.decodeBoss<PayloadBlock>() }
                    if (payloadBlock.isFailure) {
                        warning { "ignoring unknown payload block (decryption was ok)" }
                        debug { "unknown payload: \n${decryptedData.toDump()}" }
                    } else {
                        try {
                            processPayload(payloadBlock.getOrThrow())
                        } catch (t: Throwable) {
                            exception { "failed to process AsyncBlock" to t }
                        }
                    }
                }
            }
        }
    }

    private val _pushes = MutableSharedFlow<PushPayload>()

    private suspend fun processPayload(block: PayloadBlock) {
        when (block) {
            is PayloadBlock.Response -> {
                access.withLock { deferred.remove(block.toId) }?.complete(block.result)
            }
            is PayloadBlock.Push -> _pushes.emit(block.payload)

            else -> {
                warning { "not supported payload: $block" }
            }
        }
    }

    private val smPairs: MutableMap<String, SubscriptionManager> = oldChannel?.smPairs ?: mutableMapOf()

    suspend fun watchPair(cp: CurrencyPair, on: Boolean) {
        access.withLock {
            val sm = smPairs.getOrPut(cp.label) {
                SubscriptionManager(cp.label) {
                    channel().command("Watch.trader", WatchTraderArgs(cp, on))
                }
            }
            sm.subscribe(on)
        }
    }

    private val smUser: SubscriptionManager = SubscriptionManager("user", oldChannel?.smUser) { on ->
        command("Watch.user", "on" to on)
    }

    suspend fun watchUser(on: Boolean) {
        smUser.subscribe(on)
    }

    private val smLogs: SubscriptionManager = SubscriptionManager("logs", oldChannel?.smLogs) {
        rawCommand("Push.Logs.start")
    }

    suspend fun subscribeToLogs() {
        smLogs.subscribe()
    }

    private suspend fun sendBlock(block: AsyncBlock) {
        socket.send(BossEncoder.encode(block))
    }

    init {
        logLevel = Log.Level.DEBUG
        // important part: as channels are getting closed and reopened, they convey their pushes to a central:
        globalLaunch {
            _pushes.collect(globalPushes)
        }
    }

    companion object : LogTag("RTCHN") {
        private var _channel = CompletableDeferred<RealtimeChannel>()

        private val globalPushes = MutableSharedFlow<PushPayload>()


        suspend fun channel(): RealtimeChannel {
            return _channel.await()
        }

        suspend fun subscribeToLogs(): Flow<PushPayload.SystemLog> {
            channel().subscribeToLogs()
            @Suppress("UNCHECKED_CAST")
            return globalPushes.filter { it is PushPayload.SystemLog } as Flow<PushPayload.SystemLog>
        }

        suspend fun watchPair(cp: CurrencyPair, on: Boolean) {
            channel().watchPair(cp, on)
        }

        suspend fun getLastStick(cp: CurrencyPair, statSampleRate: StatSampleRate): PushPayload.JumpingCandleStick =
            BossDecoder.decodeFrom(
                channel().command("Watch.trader.getLastStick", WatchTraderLastStickArgs(cp, statSampleRate))!!
            )

        suspend fun watchUser(on: Boolean) {
            channel().watchUser(on)
        }

        suspend fun accountsFlow(): Flow<PushPayload.AccountChanged> {
            watchUser(true)
            return flow {
                globalPushes.collect { if (it is PushPayload.AccountChanged) emit(it) }
            }
        }

        val pushes: SharedFlow<PushPayload> = globalPushes

        private val _connected = MutableStateFlow<Boolean>(false)
        val connected: StateFlow<Boolean> = _connected

        private suspend fun maintainConnection() {
            var stop = false
            debug { "maintaininig realtime connection: start" }
            var channel: RealtimeChannel? = null
            while (!stop) {
                try {
                    debug { "connecting to service..." }
                    val session = net.sergeych.ugdx.GuestAPI.connection.await()
                    val url = session.rootConnection.url
                    val client = session.rootConnection.client

                    val wsUrl = when {
                        url.startsWith("https://") -> "wss" + url.substring(5)
                        url.startsWith("http://") -> "ws" + url.substring(4)
                        else -> throw IllegalArgumentException("unsupported rootUrl to connect websocket: $url")
                    }.replace("/p1", "/push1")

                    debug { "connecting websocket $wsUrl" }
                    client.webSocket(wsUrl) {
                        if( channel == null )
                            debug { "creating new channel"}
                        else
                            debug { "reconnecting existing channel"}
                        channel = RealtimeChannel(this, session, channel)
                        channel?.start { channel ->
                            debug { "WS inner connection state changed: $channel" }
                            if (channel != null) _channel.complete(channel)
                            _connected.value = true
                        }
                        debug {"Socket is closed" }
                    }
                } catch (_: CancellationException) {
                    warning { "realtime channel coroutine is cancelled, connection will not be restored" }
                    stop = true
                } catch (t: Throwable) {
                    exception { "realtime channel failed, will retry" to t }
                    delay(1000)
                    debug { "retrying connection" }
                } finally {
                    _connected.value = false
                    if( _channel.isCompleted )
                        _channel = CompletableDeferred<RealtimeChannel>()
                }
            }
        }

        init {
            logLevel = Log.Level.DEBUG
            globalLaunch { maintainConnection() }
        }
    }
}