package net.sergeych.ugdx.realtime

import net.sergeych.mp_logger.*
import net.sergeych.mp_tools.globalLaunch

/**
 * Manager for realtime channel subscriptions used to control on/off commands for multiple subscribe/unsubscribe
 * calls and to simplify channel restoration
 */
internal class SubscriptionManager(
    val name: String,
    initialCount: Int? = null,
    private val handler: suspend (Boolean) -> Unit,
) : Loggable {

    /**
     * When reconstructing channel, use it to restore subscriptions. It copies state and call handler to
     * restore subscription if need in a separate coroutine
     */
    constructor(sm: SubscriptionManager,handler: suspend (Boolean) -> Unit): this(sm.name, sm.count, handler)

    constructor(name: String,existing: SubscriptionManager?,handler: suspend (Boolean) -> Unit)
            : this(name, existing?.count, handler)

    var count: Int = initialCount ?: 0
        private set

    val isActive: Boolean  get() = count > 0

    override var logLevel: Log.Level? = Log.Level.DEBUG
    override var logTag: String = "SNM:$name"

    private suspend fun perform(on: Boolean) {
        try {
            debug { "calling handler($on)"}
            handler(on)
        }
        catch(t: Throwable) {
            exception { "failed to perform subscription handler($on)" to t }
        }
    }

    suspend fun subscribe(on: Boolean = true) {
        debug { "called subscribe($on)"}
        if( count < 0 ) warning { "called subscribe() on disconnected instance" }
        else {
            if (on)
                if (count++ == 0) perform(true)
                else
                    unsubscribe()
        }
    }

    suspend fun unsubscribe() {
        debug { "called unsubscribe()"}
        if (count > 0) {
            perform(false)
            count--
        }
    }

    init {
        if( count > 0) globalLaunch {  perform(true) }
    }
}