package net.sergeych.ugdx.shared

import net.sergeych.ugdx.DerivationParams
import kotlinx.datetime.Instant
import kotlinx.serialization.Serializable
import net.sergeych.unikrypto.*

/*
 * Inner structures that are used in API data exchange with ugdx server:
 */

@Serializable
data class AccountInfo(
    /**
     * Unique user id in the system. This is the only way to compare users as login names, public keys can be changed
     * and [createdAt] is not unique.
     */
    val id: Long,
    /**
     * Current login name. Could be used to connect between users. Could also be hidden, in which case will
     * contain empty or even a random string.
     */
    val loginName: String,
    val isAdmin: Boolean,
    /**
     * Blocked by administration. Account functionality is limited.
     */
    val isBlocked: Boolean,
    /**
     * Permission to use exchange service. Service policies control it
     */
    val canUseExchange: Boolean,
    /**
     * Test users are faster to register, but (1) can be deleted by anybody (test scripts) and are restricted in
     * operations with valuables.
     */
    val isTest: Boolean,
    /**
     * Invite bonus normally is set when an account is created with a valid invite code, reduces commission costs
     * for operations. This binus could also be awarded later, for example, if many users had used this one as
     * inviter.
     */
    val hasInviteBonus: Boolean,

    /**
     * The current public key identity.
     */
    val publicKeyId: KeyIdentity,
    /**
     * Dd not rely on it. Account creation time.
     */
    val createdAt: Instant,

    val encryptedKeyring: ByteArray,
    val derivationParams: DerivationParams,
    val partyId: String
)

/**
 * Erguments of the registration commands are in the signed record. This is mainly needed
 * to prevent logic error, as the data are transferred via parsec channel:
 */
@Serializable
data class SignedRecordArgs(
    val signedRecord: ByteArray
) {
    fun unpack(block: (SignedRecord) -> Unit ): SignedRecord = SignedRecord.unpack(signedRecord, block)

    companion object {
        inline operator fun <reified T>invoke(key: PrivateKey,payload: T,nonce: ByteArray?=null): SignedRecordArgs =
            SignedRecordArgs(SignedRecord.pack(key, payload, nonce))
    }
}


@Serializable
class RegisterArgsPayload(
    val loginName: String,
    val token: ByteArray,
    val encryptedLoginObject: ByteArray,
    val encryptedKeyring: ByteArray,
    val publicKey: PublicKey,
    val derivationParams: DerivationParams,
    val inviteCode: String?,
    val testMode: Boolean = false
)

// answer is account info or an error


@Serializable
data class SignInRequestArgs(
    val loginName: String
)
@Serializable
data class SignInRequestResult(
    val kdfParams: DerivationParams,
    val tokenSeed: ByteArray
)

@Serializable
data class SignInObjectRequestArgs(
    val loginName: String,
    val seededToken: ByteArray
)

@Serializable
data class SignInObjectRequestResult(
    val encryptedLoginObject: ByteArray
)

@Serializable
data class SignInPayload(
    val onLoginAction: String?
)
// result is account info

@Serializable
data class LoginRequest(
    val loginName: String,
    val token: ByteArray
)

/**
 * Seed token to check (used in login procedure)
 */
fun seedToken(seed: ByteArray,token: ByteArray): ByteArray = HashAlgorithm.SHA3_256.digest(seed, token, seed, token)

@Serializable
data class CheckStatusResult(
    val serviceVersion: String,
    val accountInfo: AccountInfo?,
    val sessionExpiresAt: Instant
)
